@@ -3,7 +3,8 @@ from abc import abstractmethod
 from enum import Enum
 import numpy as np
 
-# from coverage_strategies.StrategyGenerator import get_strategy_from_enum
+from coverage_strategies.coverage_strategies.Dijkstra import get_graph_from_board, dijkstra, shortest
+from coverage_strategies.coverage_strategies.SpanningTreeCoverage import is_slot_shallow_obstacle
 
 
 class Board:
@@ -27,8 +28,31 @@ class Board:
             s.is_occupied = False
             s.covered_by = "*"
 
-    def add_obstacles(self,obs=[]):
-        self.Obstacles.extend(obs)
+    def populate_with_obstacles(self, percentage: float, init_positions: list):
+        import random
+        obs = []
+
+        def get_shallow_obs(b: Board, o):
+            return [i for j in b.Slots for i in j if is_slot_shallow_obstacle(i, o)]
+
+        def is_still_connected():
+            leveled_slots = assign_level_to_slots(self, Slot(0, 0), levelType=4)
+            return self.Rows*self.Cols - len(self.get_shallow_obstacles()) == len(leveled_slots)
+            # return not any([ss not in leveled_slots and not is_slot_shallow_obstacle(ss, self.Obstacles) for ss in [i for j in self.Slots for i in j]])
+
+        while len(self.get_shallow_obstacles()) < (percentage / 100.0) * self.Rows * self.Cols:
+            o = random.choice([i for j in self.Slots for i in j
+                               if i not in self.Obstacles and
+                               i not in init_positions and
+                               0 < i.row < self.Rows - 1 and
+                               0 < i.col < self.Cols - 1])
+
+            self.Obstacles.append(o)
+            if not is_still_connected():
+                self.Obstacles.remove(o)
+
+    def get_shallow_obstacles(self):
+        return [s for sl in self.Slots for s in sl if is_slot_shallow_obstacle(s, self.Obstacles)]
 
 class Slot:
     def __init__(self, x, y):
@@ -36,7 +60,7 @@ class Slot:
         self.covered_by = "*"
         self.row = x
         self.col = y
-        self.Name = "("+str(x) + "," + str(y) + ")"
+        self.Name = "(" + str(x) + "," + str(y) + ")"
 
     def __eq__(self, other):
         return self.row == other.row and self.col == other.col
@@ -59,7 +83,10 @@ class Slot:
     def go_south(self):
         return Slot(self.row + 1, self.col)
 
-    def go(self, s, opposite_direction = False):
+    def in_bounds(self, board: Board):
+        return 0 <= self.row < board.Rows and 0 <= self.col < board.Cols
+
+    def go(self, s, opposite_direction=False):
         """
         Go South, East, North or West, according to the given parameter.
         :param s: Must by 'd', 'r','n','l'
@@ -96,6 +123,20 @@ class Slot:
     def __repr__(self):
         return str(self)
 
+    def get_inbound_neighbors(self, board: Board):
+        return [i for i in [self.go_south(), self.go_east(), self.go_west(), self.go_north()] if i.in_bounds(board)]
+
+    def get_8_inbound_neighbors(self, board: Board):
+        return [i for i in [self.go_south(),
+                            self.go_east(),
+                            self.go_west(),
+                            self.go_north(),
+                            self.go_south().go_west(),
+                            self.go_north().go_west(),
+                            self.go_south().go_east(),
+                            self.go_north().go_east()]
+                if i.in_bounds(board)]
+
 
 class StrategyEnum(Enum):
     VerticalCoverageCircular = 0
@@ -134,19 +175,20 @@ class Agent:
 
     def get_tdv(self):
         from math import fabs
-        return sum([(1/(1+stepI))*(fabs(self.steps[stepI].row - self.steps[0].row)+fabs(self.steps[stepI].col - self.steps[0].col)) for stepI in range(len(self.steps))])
+        return sum([(1 / (1 + stepI)) * (
+                    fabs(self.steps[stepI].row - self.steps[0].row) + fabs(self.steps[stepI].col - self.steps[0].col))
+                    for stepI in range(len(self.steps))])
 
     def get_frame(self):
         pass
 
-
     def get_strategy(self):
         return self.Strategy.__str__()
 
     def get_strategy_short(self):
         return self.get_strategy()[:5] + "..."
 
-    def display_heat_map(self,x,y):
+    def display_heat_map(self, x, y):
         arr = self.get_heatmap()
         # DisplayingClass.create_heat_map(arr, x, y, self.get_strategy_short())
 
@@ -182,8 +224,9 @@ class Agent:
     #     #                                                                           np.average(c),
     #     #                                                                           np.std(c)))
 
+
 class Game:
-    def __init__(self, agent_r: Agent, agent_o: Agent, size=(100,100)) -> None:
+    def __init__(self, agent_r: Agent, agent_o: Agent, size=(100, 100)) -> None:
         self._board = agent_r.gameBoard
         self._agentR = agent_r
         self._agentO = agent_o
@@ -194,46 +237,29 @@ class Game:
 
         if enforce_paths_length:
             if not len(steps_o) == len(steps_r):
-                raise AssertionError("wrong length! len(steps_o)={}, len(steps_r)={}".format(len(steps_o),len(steps_r)))
+                raise AssertionError(
+                    "wrong length! len(steps_o)={}, len(steps_r)={}".format(len(steps_o), len(steps_r)))
 
-        for i in range(min(len(steps_r), len(steps_o))):
-            # perform step for R
-            step_r = steps_r[i]
-            if not self._board.Slots[int(step_r.row)][int(step_r.col)].has_been_visited:
-                self._board.Slots[int(step_r.row)][int(step_r.col)].has_been_visited = True
-                self._board.Slots[int(step_r.row)][int(step_r.col)].covered_by = self._agentR.Name
-
-            # then perform step for O
-            step_o = steps_o[i]
-            if not self._board.Slots[int(step_o.row)][int(step_o.col)].has_been_visited:
-                self._board.Slots[int(step_o.row)][int(step_o.col)].has_been_visited = True
-                self._board.Slots[int(step_o.row)][int(step_o.col)].covered_by = self._agentO.Name
+        for curr_slot in [s for sl in self._board.Slots for s in sl
+                          if not is_slot_shallow_obstacle(s, self._agentO.gameBoard.Obstacles)]:
+            # print(self._agentO.gameBoard.Obstacles)
+            # exit(1)
+            ri = steps_r.index(curr_slot)
+            oi = steps_o.index(curr_slot)
+            curr_slot.has_been_visited = True
+            curr_slot.covered_by = self._agentO.Name if oi < ri else self._agentR.Name
 
+        # make sure all cells, which are not obstacles, are covered
+        assert all([(True if self._board.Slots[i][j].has_been_visited else False)
+                    for i in range(self._board.Rows) for j in range(self._board.Cols)
+                    if not is_slot_shallow_obstacle(Slot(i, j), self.board.Obstacles)])
         return self.get_r_gain(), self.get_o_gain()
+
     def get_r_gain(self):
-        cond_count = 0
-
-        # print self._board.Slots
-
-        for i in range(0, self._board.Rows):
-            for j in range(0, self._board.Cols):
-                if self._board.Slots[i][j].covered_by == self._agentR.Name:
-                    cond_count += 1
-
-        return float(cond_count)
+        return float(len([s for sl in self._board.Slots for s in sl if s.covered_by == self._agentR.Name]))
 
     def get_o_gain(self):
-        cond_count = 0
-
-        size_x = len(self._board.Slots)
-        size_y = len(self._board.Slots[0])
-
-        for i in range(0, size_x):
-            for j in range(0, size_y):
-                if self._board.Slots[i][j].covered_by == self._agentO.Name:
-                    cond_count += 1
-
-        return float(cond_count)
+        return float(len([s for sl in self._board.Slots for s in sl if s.covered_by == self._agentO.Name]))
 
     @property
     def board(self):
@@ -242,7 +268,7 @@ class Game:
 
 class Strategy:
     __metaclass__ = ABCMeta
-    steps = [] # type: List[Slot]
+    steps = []  # type: list[Slot]
 
     def __init__(self):
         self.steps = []
@@ -253,7 +279,7 @@ class Strategy:
 
     @classmethod
     @abstractmethod
-    def get_steps(self, agent_r, board_size = 50, agent_o = None):
+    def get_steps(self, agent_r, board_size=50, agent_o=None):
         """Returns the steps agent perform to cover the world"""
 
     @classmethod
@@ -278,6 +304,17 @@ class Strategy:
 
         return steps_to_return
 
+    @classmethod
+    def go_from_a_to_b_dijkstra(self, a, b, board):
+        g = get_graph_from_board(board)
+        dijkstra(g, g.get_vertex(a), g.get_vertex(b))
+
+        target = g.get_vertex(b)
+        path = [target.get_id()]
+        shortest(target, path)
+
+        return path[::-1]
+
     @classmethod
     def get_farthest_corner(self, a, board_size):
         """
@@ -290,3 +327,42 @@ class Strategy:
         f_col = 0 if a.col > board_size / 2 else board_size - 1
         return Slot(f_row, f_col)
 
+
+def assign_level_to_slots(board: Board, init: Slot, levelType: int = 8):
+    leveled_vertices = {}
+
+    # Mark all the vertices as not visited
+    visited = {i: False for j in board.Slots for i in j}
+    queue = [(init, 0)]
+    visited[init] = True
+    while queue:
+        s, l = queue.pop(0)
+        leveled_vertices[s] = l
+
+        # Get all adjacent vertices of the
+        # dequeued vertex s. If a adjacent
+        # has not been visited, then mark it
+        # visited and enqueue it.
+        # all unvisited neighbors get level value of +1 of the current level value
+
+        neighbors = s.get_inbound_neighbors(board) if levelType == 4 else s.get_8_inbound_neighbors(board)
+
+        for i in [i for i in neighbors if not is_slot_shallow_obstacle(i, board.Obstacles)]:
+            if not visited.get(i):
+                queue.append((i, l + 1))
+                visited[i] = True
+
+    return leveled_vertices
+
+
+def display_board(b: Board):
+    """
+    takes in a board, and saves it as a csv file.
+    :param b:
+    :return:
+    """
+    import csv
+    with open('./board.csv', 'w', newline='') as file:
+        writer = csv.writer(file)
+        shallows = b.get_shallow_obstacles()
+        writer.writerows([[(' ' if j not in shallows else '*') for j in i] for i in b.Slots])
